import http from 'k6/http'
import { textSummary } from 'https://jslib.k6.io/k6-summary/0.0.2/index.js'

import { Trend, Counter } from 'k6/metrics'

import { graphQLApiUrl } from './const.js'
import { generateRandomString, loginOryGetAuthHeaders } from './utils.js'
import { userByIdQuery } from './users/queries.js'
import { spaceByIdQuery } from './spaces/queries.js'
import { createSpaceMutation, deleteSpaceMutation } from './spaces/mutations.js'
import { inspirationPostsQuery, postByIdQuery } from './posts/queries.js'
import { insightByIdQuery } from './insights/queries.js'
import { notificationsQuery } from './common/queries.js'
import { lastInsightsQuery } from './home/queries.js'
import { createTaskMutation, deleteTaskMutation } from './tasks/mutations.js'
import { FormData } from 'https://jslib.k6.io/formdata/0.0.2/index.js'

export const options = {
  scenarios: {
    default: {
      executor: 'constant-arrival-rate',
      duration: '10s',
      // start 10 iterations/requests every ~5s
      rate: 1,
      timeUnit: '1s',
      // Pre-allocate much more VUs than we need, so that long response times don't delay the tests
      preAllocatedVUs: 30,
      gracefulStop: '300s',
    },
  },
  summaryTrendStats: ['avg', 'min', 'med', 'max', 'p(90)', 'p(95)', 'p(99)', 'count'],
}

const requestCounter = new Counter('custom_count')
const failureCounter = new Counter('custom_failure_count')
const requestDuration = new Trend('custom_duration', true)
const testToRun = __ENV.TEST

export function setup() {
  if (!(testToRun in tests)) {
    throw `don't know test ${testToRun}, choose one of ${Object.keys(tests)}`
  }
  return loginOryGetAuthHeaders()
}

// test data
const userId = 'acb1c578-8f11-4eb2-b4c3-513844e8a842' // oles user
const spaceId = 'ba8a3a58-9e98-415b-ad4d-1c68f7067cae' // A space created by Rubia named "Rogers" from May 2023, should be stable
const insightId = '9b67a7e4-a96b-4312-b5d2-7c628c462e01' // The future of mobility, published May 2023 by an external contributor
const postId = 'eb67086d-7604-4362-8fd1-f5b56e94b10a' // Post from Joao ("Y'all read this") from January 2023
const topicId = '2ebecfaa-3a6d-4f39-9389-28bbbca628c9' // Agriculture & Food
/**
 * Every map entry is a test. The key is its name, the value is its definition.
 * It can be a simple (map) object, which is then plainly serialized to the request body as JSON.
 * It can also be a FormData object, in which case the request will be sent as a multipart formdata request.
 * If the test creates objects, it can define a cleanupFunction function, which removes the created object,
 * so that there is no performance degradation due to the culmination of many objects in the database.
 */
const tests = {
  // "stable" queries on stable objects, should be comparable over time
  userById: { request: { query: userByIdQuery, variables: { userId } } },
  spaceById: { request: { query: spaceByIdQuery, variables: { idOrName: spaceId } } },
  insightById: { request: { query: insightByIdQuery, variables: { id: insightId } } },
  postById: { request: { query: postByIdQuery, variables: { postId } } },
  // "unstable" queries, not comparable over time, need to be measured before and after a change
  notifications: { request: { query: notificationsQuery, variables: { offset: 0, limit: 100 } } },
  lastInsights: { request: { query: lastInsightsQuery, variables: { offset: 0, limit: 5 } } },
  inspirationPosts: { request: { query: inspirationPostsQuery, variables: { offset: 0, limit: 10 } } },
  // mutations
  createSpace: {
    request: (() => {
      const fd = new FormData()
      fd.append(
        'operations',
        JSON.stringify({
          query: createSpaceMutation,
          variables: {
            input: {
              title: generateRandomString('perf-test-space'),
              description: 'test',
              topics: topicId,
              cover: null,
            },
          },
        }),
      )
      fd.append('map', '{ "0": ["variables.input.cover"] }')
      fd.append('0', {
        data: new Uint8Array(open('../static/spaceCover.png', 'b')).buffer,
        filename: 'cover.png',
        content_type: 'image/png',
      })
      return fd
    })(),
    cleanupFunction: (res, headers) => {
      const response = http.post(
        graphQLApiUrl,
        JSON.stringify({
          query: deleteSpaceMutation,
          variables: { id: res.data.createSpace.id },
        }),
        {
          headers,
          timeout: '300s',
        },
      )
      const body = JSON.parse(response.body)
      if (body.errors) {
        console.error(`Cleanup Request Failed: ${JSON.stringify(body.errors)}`)
      }
    },
  },
  createTask: {
    request: {
      query: createTaskMutation,
      variables: {
        input: {
          spaceId,
          visibility: 'MEMBERS',
          locationType: 'REMOTE',
          regularity: 'REGULARLY',
          contactEmail: 'test@example.org',
          description: 'test',
          name: generateRandomString('perf-test-task'),
          slotsAvailable: 5,
        },
      },
    },
    cleanupFunction: (res, headers) => {
      res = http.post(
        graphQLApiUrl,
        JSON.stringify({
          query: deleteTaskMutation,
          variables: { id: res.data.createTask.id },
        }),
        {
          headers,
          timeout: '300s',
        },
      )
      const body = JSON.parse(res.body)
      if (body.errors) {
        console.error(`Cleanup Request Failed: ${JSON.stringify(body.errors)}`)
      }
    },
  },
}

export default function (data) {
  const payload = tests[testToRun].request
  var res
  if (payload instanceof FormData) {
    res = http.post(graphQLApiUrl, payload.body(), {
      headers: Object.assign({}, data.graphQLAuthenticatedUserHeaders, {
        'Content-Type': `multipart/form-data; boundary=${payload.boundary}`,
      }),
      timeout: '300s',
    })
  } else {
    res = http.post(graphQLApiUrl, JSON.stringify(payload), {
      headers: data.graphQLAuthenticatedUserHeaders,
      timeout: '300s',
    })
  }

  var requestFailed = true
  try {
    var body = JSON.parse(res.body)
    if (res.status !== 200) {
      console.error(`Failed with HTTP status '${res.status}', response '${res.body}'`)
    } else if (body.errors) {
      console.error(`Failed in GraphQL with errors: ${JSON.stringify(body.errors)}`)
    } else {
      requestFailed = false
    }
  } catch (e) {
    console.error('Failed due to unexpected Error', e)
  }
  failureCounter.add(requestFailed)
  if (!requestFailed) {
    requestCounter.add(1)
    requestDuration.add(res.timings.duration)
    // execute cleanup function for created element
    if (typeof tests[testToRun].cleanupFunction === 'function') {
      tests[testToRun].cleanupFunction(body, data.graphQLAuthenticatedUserHeaders)
    }
  }
}

export function handleSummary(data) {
  Object.keys(data.metrics)
    .filter((key) => !key.startsWith('custom'))
    .forEach((key) => delete data.metrics[key])
  delete data.setup_data
  data.root_group.groups = []

  data.testRun = testToRun
  const result = {
    stdout: textSummary(data, { enableColors: true }),
  }
  result[`${testToRun}.json`] = JSON.stringify(data)
  return result
}
